import os
import time
import zipfile
from dnalib.log import log
from .utils import Utils

class LandingZoneUtils:

    @staticmethod   
    def load_files_dataframe(landing_path, ascending=False):
        file_path = Utils.landingzone_path(landing_path) 
        files_dataframe = None       
        try:            
            files_dataframe = (
                Utils.spark_instance().createDataFrame(Utils.dbutils().fs.ls(file_path))
                .filter("(name <> 'processados/')") 
                .select('path', 'name', 'modificationTime')
                .orderBy("modificationTime", ascending=ascending)
                .select('path','name')                
            )
        except Exception as e:
            log(__name__).error(f"Unable to load files for path {file_path}: {e}")            
            raise Exception(f"Unable to load files for path {file_path}")

        return files_dataframe

    @staticmethod
    def load_last_updated_path(landing_path):
        files_dataframe = LandingZoneUtils.load_files_dataframe(landing_path)
        return files_dataframe.first()["path"]
    


from .tables import LayerTable, ProcessLayerTable
from .silvertable import SilverTable, SilverLoaderType
from .bronzetable import BronzeTable, SourceType, LandingLoader, FileLoaderFactory, ParquetLandingLoader


from dnalib.utils import TableUtils, Utils
from dnalib.log import log

class TableComment:             
    """
        Classe construída para compilar o template do comentário da tabela, seguindo os critérios do catalogo do Atlan.

        Args:
            layer (str): camada da tabela no lake.   
            table_name (str): string que representa o nome da tabela.
            yml (yml object or str, optional): caminho do arquivo yml ou instância do objeto com os parâmetros necessários para o Create Table. Padrão é None.
            anonimized_fields (list, optional): lista de campos para serem anonimizados.
            comment (dict, optional): dicionário com o padrão table_comment_atlan_pattern.key:comment (veja TableComment.table_comment_atlan_pattern). Se não for informado, será inferido a partir do yml.
            comment_squad (dict, optional): dicionário com o padrão table_comment_atlan_pattern.key:comment_squad (veja TableComment.table_comment_atlan_pattern). Se não for informado, será inferido a partir do yml.

        Attributes:
            table_comment_atlan_pattern (dict): dicionário com as chaves para o comentário da tabela, seguindo os critérios do catalogo do Atlan.
            yml (yml object ou None): instância do objeto yml, é None caso nenhum arquivo seja informado.
            layer (str): camada da tabela no lake.   
            table_name (str): string que representa o nome da tabela.
            anonimized_fields (list): lista de campos para serem anonimizados.
            comment (dict): dicionário com o padrão table_comment_atlan_pattern.key.
            comment_squad (dict): dicionário com o padrão table_comment_atlan_pattern.key.
            parsed_comment (str): string com o template do comentário da tabela compilado, preenchido após a chamada do método parse.
            comment_complete (dict): merge entre os dicionários comment e comment_squad, preenchido após a chamada do método parse.
            parsed_template (str): template do comentário da tabela compilado, preenchido após a chamada do método parse.
    """ 

    table_comment_atlan_pattern = {
        'descricao': '- **Descrição**:',
        'sistema_origem': '- **Sistema Origem**:',
        'calendario': '- **Calendário de Disponibilização**:',
        'tecnologia': '- **Tecnologia Sistema Origem**:',
        'camada': '- **Camada de Dados**:',
        'peridiocidade': '- **Periodicidade de Atualização**:',
        'retencao': '- **Período de Retenção**:',
        'vertical': '- **Vertical**:',
        'dominio': '- **Domínio**:',
        'gi': '- **GI - Gestor da Informação**:',
        'gdn': '- **GDN - Gestor de Negócio**:',
        'curador': '- **Curador**:',
        'custodiante': '- **Custodiante**:',
        'squad': '- **Squad Responsável**:',
        'categoria_portosdm': '- **Categoria e Grupo PortoSDM**:',
        'confidencialidade': '- **Classificação da Confidencialidade**:',
        'classificacao': '- **Classificacao**:',
        'campos_anonimizados': '- **Campos Anonimizados**:'
    }

    def __init__(self, 
                 layer, 
                 table_name, 
                 yml=None, 
                 anonimized_fields=[], 
                 comment={}, 
                 comment_squad={}):
        self.yml = yml
        # yml must be either a string or a dict from the yml file
        if isinstance(self.yml, str):
            self.yml = Utils.safe_load_and_parse_yml(yml)
        self.table_name = table_name.strip().lower()
        self.layer = layer.strip().lower()   
        self.anonimized_fields = anonimized_fields
        self.comment = comment
        self.comment_squad = comment_squad
        self.parsed_comment = ""
        self.comment_complete = {}
        self.parsed_template = ""

    
    def __load_and_verify_params(self):
        """
            Método interno que carrega as estruturas principais para popular os comentários da tabela.

            Raises:
                ValueError: Ocorre se alguma das estruturas comment ou comment_squad não forem informadas. Ou devem ser passadas por parâmetros, ou devem estar dentro do yml.
        """
        
        if len(self.anonimized_fields) == 0 and self.yml != None:
            if "anonimized_fields" in self.yml:    
                self.anonimized_fields = self.yml["anonimized_fields"]

        if len(self.comment) == 0:
            if not "comment" in self.yml:
                log(__name__).error(f"If not informed, the comment parameter is required in yml table template for {self.layer} layer.")
                raise ValueError(f"If not informed, the comment parameter is required in yml table template for {self.layer} layer.")  
            self.comment = self.yml["comment"]

        if len(self.comment_squad) == 0:
            if not "comment_squad" in self.yml:
                log(__name__).error(f"If not informed, the comment_squad parameter is required in yml table template for {self.layer} layer.")
                raise ValueError(f"If not informed, the comment_squad parameter is required in yml table template for {self.layer} layer.")  
            self.comment_squad = self.yml["comment_squad"]

        # merge both dicts
        self.comment_complete = {**self.comment, **self.comment_squad}
        self.comment_complete["campos_anonimizados"] = ", ".join(self.anonimized_fields)
        
    def parse(self): 
        """
            Método que faz o parse dos comentários baseados na estrutura do Atlan. Os valores estão no dicionário TableComment.table_comment_atlan_pattern.

            Returns:
                self: a instância da classe TableComment.

            Raises:
                ValueError: Se a estrutura do comentário não for informada corretamente (ou seja, se faltar alguma chave).
        """

        self.__load_and_verify_params()                      
        for key in self.table_comment_atlan_pattern:
            if not key in self.comment_complete:
                log(__name__).error(f"You must inform the parameter {key} in table comment.")
                raise ValueError(f"You must inform the parameter {key} in table comment.")
            else:                
                self.parsed_comment += f"{self.table_comment_atlan_pattern[key]} {TableUtils.format_comment_content(self.comment_complete[key])}\n"
        return self
    
    def template(self):             
        """
            Método que constroí o template da estrutura do comentário da tabela baseado nos parâmetros informados.

            .. code-block:: SQL

                COMMENT ON TABLE layer.table_name IS "comentário compilado"
            
            Returns:
                self: a instância da classe TableComment.
        """   

        self.parse()    
        self.parsed_template = """
            COMMENT ON TABLE {}.{} IS "{}"
        """.format(self.layer, self.table_name, self.parsed_comment)           

        return self             

    def execute(self):        
        """
            Método que executa todo processo para gerar e escrever o comentário da tabela.              
            
            Returns:                
                self: a instância da classe TableComment.

            Raises:
                Exception: Se a tabela informada não existir.
        """

        # verify if table exists
        if not TableUtils.table_exists(self.layer, self.table_name):
            log(__name__).error(f"Table {self.layer}.{self.table_name} does not exist")
            raise Exception(f"Table {self.layer}.{self.table_name} does not exist")

        # generate final template
        self.template()

        # run comment on table
        Utils.spark_instance().sql(self.parsed_template)        
        return self



from pyspark.sql import SparkSession
from dnalib.utils import TableUtils, Utils
from dnalib.log import log
from .tablecomment import TableComment

class CreateTable:
    """
        Classe de alto nível para o create table. Define o método abstrato de parser para as fields. Não deve ser alocada.

        Args:      
            schema (df.schema): schema do dataframe, mesmo que df.schema.
            table_name (str): string que representa o nome da tabela.
            yml (yml object or str, optional): caminho do arquivo yml ou instância do objeto com os parâmetros necessários para o Create Table. Padrão é None.
            partition_fields (list, optional): lista de campos utilizados como partição da tabela.      
            replace (bool, optional): opção que força o replace da tabela caso ela exista. Por padrão é False.

        Attributes:
            schema (df.schema): schema do dataframe, mesmo que df.schema.
            table_name (str): string que representa o nome da tabela.
            table_path (str): caminho da tabela no lakehouse.
            yml (yml object ou None): instância do objeto yml, é None caso nenhum arquivo seja informado.
            partition_fields (list): lista de campos utilizados como partição da tabela.            
            replace (bool): opção que força o replace da tabela caso ela exista. Por padrão é False.
            has_created_table (bool): parâmetro de controle interno, se True indica que a tabela foi criada ou substituída após a chamada do método execute.
            parsed_template (str): template da criação da tabela compilado, preenchido após a chamada do método template.
    """

    def __init__(self, 
                 schema,  
                 table_name, 
                 yml = None,                  
                 partition_fields = [],
                 replace = False):        
        self.table_name = table_name.strip().lower()
        self.schema = schema
        self.table_path = Utils.lakehouse_path(self.layer, self.table_name) 
        self.yml = yml
        self.partition_fields = partition_fields        
        self.replace = replace 
        if isinstance(self.yml, str):
            self.yml = Utils.safe_load_and_parse_yml(yml)
        self.has_created_table = False         
        self.parsed_template = ""

    def parse_fields(self):
        """
            Método abstrato para ser implementado, faz o trabalho de parsear as fields para gerar o template do Create Table.
        """
        return NotImplementedError("Method parse_fields must be implemented")
    
    def template(self):        
        """
            Método que gera o template do Create Table e injeta no objeto self.template.
            
            .. code-block:: SQL

                CREATE OR REPLACE TABLE layer.table (
                    field_name field_type COMMENT '' nullable or not,
                    ...
                )
                USING DELTA
                PARTITIONED BY (partition_fields)

           Returns:
                self: a instância da classe CreateTable.
        """        
        # parse fields to string
        self.parse_fields()        
        # parse table partition_fields to a string
        partition_fields =  ""
        if len(self.partition_fields) > 0:
            partition_fields = ", ".join(self.partition_fields)
            partition_fields = f"PARTITIONED BY ({partition_fields})"
        # generate template for create table
        self.parsed_template = """
            CREATE OR REPLACE TABLE {}.{} (
                {}
            )
            USING delta
            {}
            LOCATION '{}'
        """.format(self.layer, self.table_name, self.parsed_fields, partition_fields, self.table_path)
        return self    

    def execute(self):  
        """
            Método que executa o create table com base no template gerado. Caso a tabela já exista, não é executado nada a não ser que o parâmetro replace=True.

            Returns:
                self: a instância da classe CreateTable.
        """        
        if not TableUtils.table_exists(self.layer, self.table_name) or self.replace:                 
            ## generate final create table template
            self.template() 
            Utils.spark_instance().sql(self.parsed_template)                                            
            # mark table has been created or replaced by class
            self.has_created_table = True    
        else:
            log(__name__).warning(f"The table already exists, so nothing will be done. To replace the table set the parameter replace=True.")
        return self

class CreateTableBronze(CreateTable):
    """
        Classe que implementa o create table para a camada bronze, é o processo mais simples de todos.

        Args:      
            schema (df.schema): schema do dataframe, mesmo que df.schema.
            table_name (str): string que representa o nome da tabela.
            yml (yml object, string or None): caminho do arquivo yml ou instância do objeto com os parâmetros necessários para o Create Table. Padrão é None.
            partition_fields (list): lista de campos utilizados como partição da tabela.
            replace (bool): opção que força o replace da tabela caso ela exista. Por padrão é False.
    """

    layer = "bronze"

    def __init__(self, schema, table_name, yml=None, partition_fields=[], replace=False):        
        super().__init__(schema, table_name, yml, partition_fields, replace)
    
    def __load_and_verify_params(self):
        if len(self.partition_fields) == 0 and self.yml != None:
            if "partition_fields" in self.yml:    
                self.partition_fields = self.yml["partition_fields"]

        if not "dataCarga" in self.schema.names:
            log(__name__).error("Any bronze table must have a dataCarga field.")
            raise Exception("Any bronze table must have a dataCarga field.")

    def parse_fields(self): 
        """
            Método que faz o parse dos campos para o create table na bronze. O método faz a leitura de cada field e gera o seguinte padrão:
            
             .. code-block:: SQL

                field_name field_type nullable or not           
            
            Returns:
                self: a instância da classe CreateTable.
        """
        self.__load_and_verify_params()      
        list_of_parsed_fields = ["{} {} {}".format(field.name, field.dataType.simpleString(), "NOT NULL" if not field.nullable else "") for field in self.schema]
        self.parsed_fields = ", ".join(list_of_parsed_fields)
        return self
    
class CreateTableSilver(CreateTable):
    """
        Classe que implementa um create table para a camada silver, baseado nos padrões do catalogo de dados do Atlan.
        
        Args:
            schema (df.schema): schema do dataframe, mesmo que df.schema.
            table_name (str): string que representa o nome da tabela.
            yml (yml object, string or None): caminho do arquivo yml ou instância do objeto com os parâmetros necessários para o Create Table. Padrão é None.
            partition_fields (list): lista de campos utilizados como partição da tabela.      
            anonimized_fields (list, optional): lista de campos para serem anonimizados.
            comment (dict, optional): dicionário com o padrão table_comment_atlan_pattern.key:comment (veja TableComment.table_comment_atlan_pattern). Se não for informado, será inferido a partir do yml.
            comment_squad (dict, optional): dicionário com o padrão table_comment_atlan_pattern.key:comment_squad (veja TableComment.table_comment_atlan_pattern). Se não for informado, será inferido a partir do yml.
            fields (dict, optional): dicionário com o padrão field_name:comment. Se não for informado, será inferido a partir do yml.
            replace (bool, optional): opção que força o replace da tabela caso ela exista. Por padrão é False.

        Attributes:
            layer (str): constante com o valor "silver".
            schema (df.schema): schema do dataframe, mesmo que df.schema.
            table_name (str): string que representa o nome da tabela.
            table_path (str): caminho da tabela no lakehouse.
            yml (yml object ou None): instância do objeto yml, é None caso nenhum arquivo seja informado.
            partition_fields (list): lista de campos utilizados como partição da tabela.            
            replace (bool): opção que força o replace da tabela caso ela exista. Por padrão é False.
            has_created_table (bool): parâmetro de controle interno, se True indica que a tabela foi criada ou substituída após a chamada do método execute.
            template (str): template da criação da tabela compilado, preenchido após a chamada do método template().
            fields (dict): dicionário com o padrão field_name:comment. 
            tbl_comment (TableComment): instância da classe TableComment para gerar o comentário da tabela.
            parsed_fields (str): string com os comentários das fields parseados no padrão do create table, preenchido após a chamada do método parse_fields().
    """

    layer = "silver"
    
    def __init__(self, 
                 schema, 
                 table_name, 
                 yml=None, 
                 partition_fields=[],
                 anonimized_fields=[], 
                 comment={}, 
                 comment_squad={}, 
                 fields={},
                 replace=False):        
        super().__init__(schema, table_name, yml, partition_fields, replace)
        self.fields = fields
        self.tbl_comment = TableComment(self.layer, self.table_name, self.yml, anonimized_fields, comment, comment_squad)
        self.parsed_fields = ""

    def __load_and_verify_params(self):
        """
            Método interno que executa a verificação dos parâmetros necessários para o create table ser executado.

            Raises:
                ValueError: Caso os parâmetro fields não seja informado; Se nem os campos do dicionário fields possuem um comentário; Se a tabela não tem o campo dataCarga.
        """     

        # it is not a required parameter, but if it is not informed, it will be overwrited by the yml parameter
        if len(self.partition_fields) == 0 and self.yml != None:
            if "partition_fields" in self.yml:                
                self.partition_fields = self.yml["partition_fields"]

        # comments are required in silver layer
        if len(self.fields) == 0:
            if not "fields" in self.yml:
                log(__name__).error(f"If not informed, the fields parameter is required in yml table template for {self.layer} layer.")
                raise ValueError(f"If not informed, the fields parameter is required in yml table template for {self.layer} layer.")  
            self.fields = self.yml["fields"]
        
        # verify if each field has its comment
        if not all(self.fields.values()):
            log(__name__).error("All fields muts have a comment.")
            raise ValueError("All fields muts have a comment.") 

        # we manually insert dataCarga to the object (if it is passed it will be generated in comments)
        if not "dataCarga" in self.schema.names:
            log(__name__).error("Any silver table must have a dataCarga field.")
            raise ValueError("Any silver table must have a dataCarga field.")
        else:
            #self.fields["dataCarga"] = "Data de carga dos dados."
            self.fields["dataCarga"] = ["Data de carga dos dados.", 'timestamp', None, None]

        # verify if checksum is added to the silver table
        if "checksum" in self.schema.names:
            #self.fields["checksum"] = "CheckSum/Hash das informações do registro (linha)."
            self.fields["checksum"] = ["CheckSum/Hash das informações do registro (linha).", 'string', None, None]

    def parse_fields(self): 
        """
            Método que faz o parse dos campos para o create table na silver. O método faz a leitura de cada field e gera o seguinte padrão:
            
             .. code-block:: SQL
            
                field_name field_type COMMENT 'field_comment' nullable or not         
            
            Returns:
                self: a instância da classe CreateTable.
        """     
        self.__load_and_verify_params()
        list_of_parsed_fields = ["{} {} COMMENT '{}' {}".format(field.name, 
                                                                field.dataType.simpleString(), 
                                                                TableUtils.format_comment_content(self.fields[field.name][0]), 
                                                                "NOT NULL" if not field.nullable else "") for field in self.schema]
        self.parsed_fields = ", ".join(list_of_parsed_fields)
        return self
        
    def execute(self):    
        """
            Método que executa o create table na camada silver e faz a chamada do comment on table, respectivamente. O comentário é gerado novamente, caso a tabela seja criada ou recriada.

            Returns:
                self: a instância da classe CreateTable.
        """      
        super().execute()
        # it forces comment on table when table is recreated
        if self.has_created_table:
            self.tbl_comment.execute()
        return self
    
# same as silver
class CreateTableGold(CreateTableSilver):
    """
        Classe que implementa um create table para a camada gold, baseado nos padrões do catalogo de dados do Atlan.
        
        Args:
            schema (df.schema): schema do dataframe, mesmo que df.schema.
            table_name (str): string que representa o nome da tabela.
            yml (yml object, string or None): caminho do arquivo yml ou instância do objeto com os parâmetros necessários para o Create Table. Padrão é None.
            partition_fields (list): lista de campos utilizados como partição da tabela.      
            anonimized_fields (list, optional): lista de campos para serem anonimizados.
            comment (dict, optional): dicionário com o padrão table_comment_atlan_pattern.key:comment (veja TableComment.table_comment_atlan_pattern). Se não for informado, será inferido a partir do yml.
            comment_squad (dict, optional): dicionário com o padrão table_comment_atlan_pattern.key:comment_squad (veja TableComment.table_comment_atlan_pattern). Se não for informado, será inferido a partir do yml.
            fields (dict, optional): dicionário com o padrão field_name:comment. Se não for informado, será inferido a partir do yml.
            replace (bool, optional): opção que força o replace da tabela caso ela exista. Por padrão é False.

        Attributes:
            layer (str): constante com o valor "gold".
            schema (df.schema): schema do dataframe, mesmo que df.schema.
            table_name (str): string que representa o nome da tabela.
            table_path (str): caminho da tabela no lakehouse.
            yml (yml object ou None): instância do objeto yml, é None caso nenhum arquivo seja informado.
            partition_fields (list): lista de campos utilizados como partição da tabela.            
            replace (bool): opção que força o replace da tabela caso ela exista. Por padrão é False.
            has_created_table (bool): parâmetro de controle interno, se True indica que a tabela foi criada ou substituída após a chamada do método execute.
            template (str): template da criação da tabela compilado, preenchido após a chamada do método template().
            fields (dict): dicionário com o padrão field_name:comment. 
            tbl_comment (TableComment): instância da classe TableComment para gerar o comentário da tabela.
            parsed_fields (str): string com os comentários das fields parseados no padão do create table, preenchido após a chamada do método parse_fields().
    """

    layer = "gold"

    def __init__(self, 
                 schema, 
                 table_name, 
                 yml=None, 
                 partition_fields=[],
                 anonimized_fields=[], 
                 comment={}, 
                 comment_squad={}, 
                 fields={},
                 replace=False):  
        super().__init__(schema, table_name, yml, partition_fields, anonimized_fields, comment, comment_squad, fields, replace)

# same as silver
class CreateTableDiamond(CreateTableSilver):
    """
        Classe que implementa um create table para a camada diamond, baseado nos padrões do catalogo de dados do Atlan.
        
        Args:
            schema (df.schema): schema do dataframe, mesmo que df.schema.
            table_name (str): string que representa o nome da tabela.
            yml (yml object, string or None): caminho do arquivo yml ou instância do objeto com os parâmetros necessários para o Create Table. Padrão é None.
            partition_fields (list): lista de campos utilizados como partição da tabela.      
            anonimized_fields (list, optional): lista de campos para serem anonimizados.
            comment (dict, optional): dicionário com o padrão table_comment_atlan_pattern.key:comment (veja TableComment.table_comment_atlan_pattern). Se não for informado, será inferido a partir do yml.
            comment_squad (dict, optional): dicionário com o padrão table_comment_atlan_pattern.key:comment_squad (veja TableComment.table_comment_atlan_pattern). Se não for informado, será inferido a partir do yml.
            fields (dict, optional): dicionário com o padrão field_name:comment. Se não for informado, será inferido a partir do yml.
            replace (bool, optional): opção que força o replace da tabela caso ela exista. Por padrão é False.

        Attributes:
            layer (str): constante com o valor "diamond".
            schema (df.schema): schema do dataframe, mesmo que df.schema.
            table_name (str): string que representa o nome da tabela.
            table_path (str): caminho da tabela no lakehouse.
            yml (yml object ou None): instância do objeto yml, é None caso nenhum arquivo seja informado.
            partition_fields (list): lista de campos utilizados como partição da tabela.            
            replace (bool): opção que força o replace da tabela caso ela exista. Por padrão é False.
            has_created_table (bool): parâmetro de controle interno, se True indica que a tabela foi criada ou substituída após a chamada do método execute.
            template (str): template da criação da tabela compilado, preenchido após a chamada do método template().
            fields (dict): dicionário com o padrão field_name:comment. 
            tbl_comment (TableComment): instância da classe TableComment para gerar o comentário da tabela.
            parsed_fields (str): string com os comentários das fields parseados no padão do create table, preenchido após a chamada do método parse_fields().
    """

    layer = "diamond"

    def __init__(self, 
                 schema, 
                 table_name, 
                 yml=None, 
                 partition_fields=[],
                 anonimized_fields=[], 
                 comment={}, 
                 comment_squad={}, 
                 fields={},
                 replace=False):  
        super().__init__(schema, table_name, yml, partition_fields, anonimized_fields, comment, comment_squad, fields, replace)

# same as silver
class CreateTableExport(CreateTableSilver):
    """
        Classe que implementa um create table para a camada export, baseado nos padrões do catalogo de dados do Atlan.
        
        Args:
            schema (df.schema): schema do dataframe, mesmo que df.schema.
            table_name (str): string que representa o nome da tabela.
            yml (yml object, string or None): caminho do arquivo yml ou instância do objeto com os parâmetros necessários para o Create Table. Padrão é None.
            partition_fields (list): lista de campos utilizados como partição da tabela.      
            anonimized_fields (list, optional): lista de campos para serem anonimizados.
            comment (dict, optional): dicionário com o padrão table_comment_atlan_pattern.key:comment (veja TableComment.table_comment_atlan_pattern). Se não for informado, será inferido a partir do yml.
            comment_squad (dict, optional): dicionário com o padrão table_comment_atlan_pattern.key:comment_squad (veja TableComment.table_comment_atlan_pattern). Se não for informado, será inferido a partir do yml.
            fields (dict, optional): dicionário com o padrão field_name:comment. Se não for informado, será inferido a partir do yml.
            replace (bool, optional): opção que força o replace da tabela caso ela exista. Por padrão é False.

        Attributes:
            layer (str): constante com o valor "export".
            schema (df.schema): schema do dataframe, mesmo que df.schema.
            table_name (str): string que representa o nome da tabela.
            table_path (str): caminho da tabela no lakehouse.
            yml (yml object ou None): instância do objeto yml, é None caso nenhum arquivo seja informado.
            partition_fields (list): lista de campos utilizados como partição da tabela.            
            replace (bool): opção que força o replace da tabela caso ela exista. Por padrão é False.
            has_created_table (bool): parâmetro de controle interno, se True indica que a tabela foi criada ou substituída após a chamada do método execute.
            template (str): template da criação da tabela compilado, preenchido após a chamada do método template().
            fields (dict): dicionário com o padrão field_name:comment. 
            tbl_comment (TableComment): instância da classe TableComment para gerar o comentário da tabela.
            parsed_fields (str): string com os comentários das fields parseados no padão do create table, preenchido após a chamada do método parse_fields().
    """

    layer = "export"

    def __init__(self, 
                 schema, 
                 table_name, 
                 yml=None, 
                 partition_fields=[],
                 anonimized_fields=[], 
                 comment={}, 
                 comment_squad={}, 
                 fields={},
                 replace=False):  
        super().__init__(schema, table_name, yml, partition_fields, anonimized_fields, comment, comment_squad, fields, replace)



from dnalib.utils import Utils, LandingZoneUtils, TableUtils
from dnalib.custom import YamlLoader
from dnalib.log import *
from dnalib.core import CreateTableBronze
from dnalib.writer import TableWriter, WriteModes
from .tables import LayerTable
from enum import Enum
from pyspark.sql.functions import from_utc_timestamp, current_timestamp, concat, col, row_number, when
from delta.tables import DeltaTable
from pyspark.sql.window import Window
import yaml

class SourceType(Enum):
    TXT = "txt"
    CSV = "csv"
    PARQUET = "parquet"

class LandingLoader(Enum):
    INCREMENTAL = 1 # read a single part, and merge data
    FULL = 2 # read all, then overwrite    
    CDC = 3 # only for cdc loader

class FileLandingLoader:

    def __init__(self, 
                 landing_files_path, 
                 table_name,
                 loader_type,
                 table_sufix,
                 name_filter = ""):
        self.landing_files_path = landing_files_path
        self.table_name = table_name
        self.loader_type = loader_type
        self.table_sufix = table_sufix
        self.name_filter = name_filter
        self.df_landing = None

    def load_incremental(self):
        # TODO: implementar o load do tipo incremental. (deve retornar um DataFrame Spark)
        raise NotImplementedError("Method load_incremental() must be implemented.")

    def load_full(self):
        # TODO: implementar o load do tipo full. (deve retornar um DataFrame Spark)
        raise NotImplementedError("Method load_full() must be implemented.")

    def load_cdc(self):
        # TODO: implementar o load do tipo cdc. (deve retornar um DataFrame Spark)
        raise NotImplementedError("Method load_cdc() must be implemented.")

    def load(self):
        if self.loader_type == LandingLoader.INCREMENTAL:
            self.df_landing = self.load_incremental()
        elif self.loader_type == LandingLoader.FULL:
            self.df_landing = self.load_full()
        elif self.loader_type == LandingLoader.CDC:
            self.df_landing = self.load_cdc()
        else:
            log(__name__).error(f"Invalid load type {self.loader_type}")
            raise ValueError(f"Invalid load type {self.loader_type}")
        return self

class ParquetLandingLoader(FileLandingLoader):

    # This implementation is based on adf:
        # table/date/ -> incremental path
        # table/date/ -> cdc path
        # table/date/ -> full path = if process run in incremental mode, but you specified a full query
        # table/ -> full path
    
    def __init__(self, landing_files_path, table_name, loader_type, table_sufix, table_config):
        super().__init__(landing_files_path, table_name, loader_type, table_sufix) 
        self.spark = Utils.spark_instance()  # Obtém a instância do Spark usando Utils.spark_instance()                        
        self.table_config = table_config

    def __load_df_parquet(self):
        # le o path mais recente dentro da pasta indicada por landing_files_path
        file_path = LandingZoneUtils.load_last_updated_path(self.landing_files_path)
        # Lê os dados do parquet atual
        try:
            df = self.spark.read.parquet(file_path) 
        except Exception as e:
            log(__name__).error(f"Error reading parquet file: {e}")          
            raise Exception(f"Error reading parquet file: {e}")
        return df

    def load_incremental(self):       
        return self.__load_df_parquet()

    def load_full(self):
        return self.__load_df_parquet()

    def load_cdc(self):
        df = self.__load_df_parquet()
        
        # Renomeia as colunas específicas para facilitar o entendimento
        df = df.withColumnRenamed("__$start_lsn", "start_lsn") \
               .withColumnRenamed("__$end_lsn", "end_lsn") \
               .withColumnRenamed("__$seqval", "seqval") \
               .withColumnRenamed("__$operation", "operation") \
               .withColumnRenamed("__$update_mask", "update_mask") \
               .withColumnRenamed("__$command_id", "command_id")
        
        (df.createOrReplaceTempView("vw_table_logs"))

        self.spark.sql(f"""
            CREATE OR REPLACE TEMP VIEW vw_table_logs_rownumber AS (
                    SELECT * FROM (
                            SELECT 
                            *, 
                            ROW_NUMBER() OVER(PARTITION BY {self.table_config['key']} ORDER BY data_alteracao DESC) AS LAST_UPDATE 
                            FROM 
                            vw_table_logs
                        )
                    WHERE LAST_UPDATE = 1
            );
        """)

        # Filtra as operações de insert e update
        df_upinsert = self.spark.sql(""" 
            SELECT * 
            FROM vw_table_logs_rownumber 
            WHERE operation IN (2, 3, 4);
            """).drop("LAST_UPDATE", "start_lsn", "end_lsn", "seqval", "operation", "update_mask", "command_id", "data_alteracao")        
        log(__name__).info(f"Load of CDC for {self.table_name} completed successfully.")
        return df_upinsert
        
class FileLoaderFactory:
    @staticmethod
    def create_file_loader(source_type, landing_files_path, table_name, loader_type, table_sufix, table_config):
        if source_type == SourceType.PARQUET:
            return ParquetLandingLoader(landing_files_path, table_name, loader_type, table_sufix, table_config)
        else:
            log(__name__).error(f"file_loader could not be created for source_type {source_type}")
            raise ValueError(f"file_loader could not be created for source_type {source_type}")
        
class BronzeTable(LayerTable):

    layer = "bronze"

    def __init__(self,                  
                 table_name,                 
                 landing_files_path,                                   
                 yaml_config_path,
                 table_sufix = "",
                 source_type=SourceType.PARQUET, 
                 loader_type=LandingLoader.FULL,
                 source_df=None, 
                 name_filter="", 
                 unzip=True, 
                 include_checksum=False):
        super().__init__(self.layer, f"{table_sufix}_{table_name}", source_df, include_checksum)        
        self.landing_files_path = landing_files_path
        self.name_filter = name_filter
        self.table_sufix = table_sufix
        self.source_type = source_type
        self.yaml_config_path = yaml_config_path
        self.table_config = Utils.yaml_table_parameters(table_name=f"{table_sufix}_{table_name}", yml_file_path=yaml_config_path)
        self.loader_type = loader_type
        self.unzip = unzip        
        self.except_fields = {}

        # Usa a fábrica para criar o file_loader correto
        self.file_loader = FileLoaderFactory.create_file_loader(
            self.source_type,
            self.landing_files_path,
            self.table_name,
            self.loader_type,
            self.table_sufix,
            self.table_config
        )        

    def parse_df_source(self):
        """
            Método que executa o carregamento de dados da camada landing para a bronze

            Returns:
                source_df (spark DataFrame): dataframe carregado a partir da camada source, caso source_df seja None.

        """         
        # load strategy in this case is to check if the table exists        
        return self.file_loader.load().df_landing

    def load_fields_from_source(self):                
        return "*"
    
    def create_table(self, yml=None, partition_fields=[]):
        if not self.has_loaded:
            self.load()
            log(__name__).warning(f"The load() method will be called internally because you call create_table first.") 
        self.creat_tbl = CreateTableBronze(self.target_df.schema, self.table_name, yml, partition_fields).execute()
        return self
    
    # implementar lógica usando mode = None (nesse caso a decisão usa o LandingLoader como parâmetro caso contrário fica a criterio de quem escolheu)
    def persist(self, mode=WriteModes.OVERWRITE, partition_fields=[], optimize=True, source_df_name="source", update_df_name="update", merge_condition=None):        
        has_delete = False                
        # when you have cdc ingestion, you may need to upsert
        if self.loader_type == LandingLoader.CDC:
            mode = WriteModes.UPSERT
            merge_condition = self.table_config["merge"]
            has_delete = True        
        elif self.loader_type == LandingLoader.INCREMENTAL:
            mode = WriteModes.UPSERT
            merge_condition = self.table_config["merge"]        

        # calling super persist method
        super().persist(mode=mode, partition_fields=partition_fields, optimize=optimize, source_df_name=source_df_name, update_df_name=update_df_name, merge_condition=merge_condition)

        if has_delete:            
            key_columns = ",".join(self.table_config["key"]) if isinstance(self.table_config["key"], list) else self.table_config["key"]
            self.writer.delete(key_columns=key_columns)




from dnalib.log import log
from dnalib.utils import Utils
from delta.tables import DeltaTable
from enum import Enum

class WriteModes(Enum):
    """ Classe com os modos de escritas válidos: overwrite, append e upsert. """
    OVERWRITE = "overwrite"    
    APPEND = "append"    
    UPSERT = "upsert"

class TableWriter:
    """ 
        Uma classe que implementa os três modos mais simples de escrita de dados: overwrite (sobreescrita completa), append (equivalente ao insert) e upsert (insert e update). 

        Args:
            table_name (str): string que representa o nome da tabela.
            layer (str): camada da tabela no lake.   
    """

    def __init__(self, table_name, layer):
        """            
            Construtor da classe. 
        """
        self.table_name = table_name
        self.layer = layer        

    def __upsert(self, df_update, merge_condition, source_df_name="source", update_df_name="update", has_checksum_field=False):
        """
            Método interno que executa o upsert.
                
            Args:   
                merge_condition (str): uma condição valida para fazer merge entre o "source_df_name" e o "update_df_name".
                has_checksum_field (bool): uma flag que indica se a tabela tem ou não o campo de checksum.            
        """
        df_source = DeltaTable.forName(Utils.spark_instance(), f"{self.layer}.{self.table_name}")
        # it improves merge performance in write operations
        update_condition = None
        if has_checksum_field:
            update_condition = f"nvl({source_df_name}.checksum, '') != {update_df_name}.checksum"
        # runing a normal merge operation
        (df_source.alias(source_df_name)
            .merge(
                df_update.alias(update_df_name),
                merge_condition
            )
            .whenMatchedUpdateAll(condition=update_condition)
            .whenNotMatchedInsertAll()
            .execute())
        
    def delete(self, key_columns):
        """
        Método que executa a exclusão de dados da tabela com base em uma condição.

        Args:
            key_columns (str): Chaves para realizar o delete.
        
        Returns:
            self: uma instância da classe TableWriter.
        """
            
        delete_condition = f"SELECT concat({key_columns}) FROM vw_table_logs_rownumber WHERE operation = 1"
        try:
            # Referencia a tabela Delta
            df_delete = Utils.spark_instance().sql(f"""DELETE FROM  {self.layer}.{self.table_name}
                           WHERE concat({key_columns}) IN ({delete_condition}) """)                        
            num_affected_rows = df_delete.first()["num_affected_rows"]
            log(__name__).info(f"{num_affected_rows} records where deleted from {self.layer}.{self.table_name}.")
        except Exception as e:
            log(__name__).error(f"Error while trying to delete records from {self.layer}.{self.table_name}: {e}")
            raise

        return self
    
    def __optimize(self):        
        """ 
            Método interno para executar o OPTIMIZE.
        """
        Utils.spark_instance().sql(f"OPTIMIZE {self.layer}.{self.table_name}")

    def persist(self, df, mode=WriteModes.OVERWRITE, partition_fields=[], optimize=True, source_df_name="source", update_df_name="update", merge_condition=None, has_checksum_field=False):              
        """
            Método que implementa escrita de dados no lake, a partir de um dataframe. Aceita um dos três modos: overwrite, append ou upsert.
                
            Args:  
                df (Spark Dataframe): dataframe Spark que vai ser persistido na tabela passada em table_name.
                mode (WriteModes): modo valido de escrita, por padrão é mode = C.OVERWRITE.
                partition_fields (list): uma lista de campos utilizadas para particionar o dataframe antes da escrita, utilizado apenas se mode != WriteModes.UPSERT.
                optimize (bool): se verdadeiro, aplica OPTIMIZE na tabela após a escrita.
                source_df_name (str): um alias para o dataframe da origem (usado apenas se o mode = WriteModes.UPSERT).
                update_df_name (str): um alias para o dataframe de update (usado apenas se o mode = WriteModes.UPSERT).
                merge_condition (str): uma condiçao valida para merge (note que source_df_name = "source" e update_df_name = "updates").
                has_checksum_field (bool): uma flag que indica se a tabela tem ou não o campo de checksum. Se verdadeiro é usado no método whenMatchedUpdateAll, isso implica que apenas registros que mudaram serão realmente reescritos nos arquivos delta.        

            Returns:
                self: uma instância da classe TableWriter.    
        """
        # it prevents to overwrite table if no data is provided
        if df.count() == 0:                                        
            log(__name__).warning(f"No data to persist in {self.table_name}, so nothing will be done.")       
        else:              
            # overwrite or append has same sintax
            if mode.value == WriteModes.OVERWRITE.value or mode.value == WriteModes.APPEND.value:         
                # load df and mode
                df_write = df.write.format("delta").mode(mode.value)
                if mode.value == WriteModes.OVERWRITE.value:
                    df_write = df_write.option("overwriteSchema", "True")
                # verify partitions
                if len(partition_fields) > 0:
                    df_write = df_write.partitionBy(partition_fields)          
                # saving data in delta table                                
                df_write.saveAsTable(f"{self.layer}.{self.table_name}")        
            elif mode.value == WriteModes.UPSERT.value:
                if merge_condition is None:
                    log(__name__).error(f"The merge_condition parameter is required for upsert mode")           
                    raise Exception(f"The merge_condition parameter is required for upsert mode")
                self.__upsert(df, merge_condition)
            else:
                log(__name__).error(f"The mode parameter {mode} is not valid. Valid values are: {[m.value for m in WriteModes]}")            
                raise Exception(f"The mode parameter {mode} is not valid. Valid values are: {[m.value for m in WriteModes]}")
            # runing optimize operation
            if optimize:            
                self.__optimize()
        return self


from .createtable import CreateTable, CreateTableBronze, CreateTableSilver, CreateTableGold, CreateTableDiamond, CreateTableExport
from .tablecomment import TableComment
from .createview import CreateView